from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import time
import trax_py3.server
import trax_py3.region
import trax_py3.image
from trax_py3.region import Rectangle
import logging
logging.getLogger().setLevel(logging.INFO)
import os
import os.path as osp
import tensorflow as tf
import numpy as np
import cv2
from cv2 import imwrite

sys.path.append('/data/anfeng/tracking/code/SA-SIAM')
LOG_DIR = '/data/anfeng/tracking/code/SA-SIAM/Logs/SA-Siam'
from utils.infer_utils import convert_bbox_format
from utils.misc_utils import get_center, get, auto_select_gpu, load_cfgs
from inference import inference_wrapper

from utils.infer_utils import get_saver

class Tracker(object):
  """Tracker based on the siamese model."""

  def __init__(self, siamese_model, model_config, track_config, sess):
    self.siamese_model = siamese_model
    self.model_config = model_config
    self.track_config = track_config

    self.num_scales = track_config['num_scales']
    logging.info('track num scales -- {}'.format(self.num_scales))
    scales = np.arange(self.num_scales) - get_center(self.num_scales)
    self.search_factors = [self.track_config['scale_step'] ** x for x in scales]

    self.x_image_size = track_config['x_image_size']  # Search image size
    self.window = None  # Cosine window
    self.log_level = track_config['log_level']

    self.sess = sess

  def init_tracker(self, first_bbox, first_frame):
    # Get initial target bounding box and convert to center based
    bbox_left, bbox_up, bbox_height, bbox_width = first_bbox.x, first_bbox.y, first_bbox.height, first_bbox.width
    bbox_cx = bbox_left + get_center(bbox_width)
    bbox_cy = bbox_up + get_center(bbox_height)

    # Feed in the first frame image to set initial state.
    bbox_feed = [bbox_cy, bbox_cx, bbox_height, bbox_width]
    input_feed = [str(first_frame), bbox_feed]
    frame2crop_scale = self.siamese_model.initialize(self.sess, input_feed)
    self.cur_bbox = first_bbox
    self.original_target_width = bbox_width
    self.original_target_height = bbox_height
    return first_bbox

  def step_tracker(self, pre_frame):
    bbox_left, bbox_up, bbox_height, bbox_width = self.cur_bbox.x, self.cur_bbox.y, self.cur_bbox.height, self.cur_bbox.width
    bbox_cx = bbox_left + get_center(bbox_width)
    bbox_cy = bbox_up + get_center(bbox_height)

    
    bbox_feed = [bbox_cy, bbox_cx, bbox_height, bbox_width]
    input_feed = [str(pre_frame), bbox_feed]

    outputs, metadata = self.siamese_model.inference_step(self.sess, input_feed)
    search_scale_list = outputs['scale_xs']
    response = outputs['response']
    response_size = response.shape[1]

    # Choose the scale whole response map has the highest peak
    if self.num_scales > 1:
      response_max = np.max(response, axis=(1, 2))
      penalties = self.track_config['scale_penalty'] * np.ones((self.num_scales))
      current_scale_idx = int(get_center(self.num_scales))
      penalties[current_scale_idx] = 1.0
      response_penalized = response_max * penalties
      best_scale = np.argmax(response_penalized)
      if np.max(response_max)<0:
        logging.warning('MAX_RESPONSE LESS THAN ZERO!')
        # best_scale = current_scale_idx
    else:
      best_scale = 0

    response = response[best_scale]

    with np.errstate(all='raise'):  # Raise error if something goes wrong
      response = response - np.min(response)
      response = response / np.sum(response)

    if self.window is None:
      window = np.dot(np.expand_dims(np.hanning(response_size), 1),
                      np.expand_dims(np.hanning(response_size), 0))
      self.window = window / np.sum(window)  # normalize window
    window_influence = self.track_config['window_influence']
    response = (1 - window_influence) * response + window_influence * self.window

    # Find maximum response
    r_max, c_max = np.unravel_index(response.argmax(),
                                    response.shape)

    # Convert from crop-relative coordinates to frame coordinates
    p_coor = np.array([r_max, c_max])
    # displacement from the center in instance final representation ...
    disp_instance_final = p_coor - get_center(response_size)
    # ... in instance feature space ...
    upsample_factor = self.track_config['upsample_factor']
    disp_instance_feat = disp_instance_final / upsample_factor
    # ... Avoid empty position ...
    r_radius = int(response_size / upsample_factor / 2)
    disp_instance_feat = np.maximum(np.minimum(disp_instance_feat, r_radius), -r_radius)
    # ... in instance input ...
    disp_instance_input = disp_instance_feat * self.model_config['embed_config']['stride']
    # ... in instance original crop (in frame coordinates)
    disp_instance_frame = disp_instance_input / search_scale_list[best_scale]
    # Position within frame in frame coordinates
    y = bbox_cy
    x = bbox_cx
    y += disp_instance_frame[0]
    x += disp_instance_frame[1]

    # Target scale damping and saturation
    target_scale = self.cur_bbox.height / self.original_target_height
    search_factor = self.search_factors[best_scale]
    scale_damp = self.track_config['scale_damp']  # damping factor for scale update
    target_scale *= ((1 - scale_damp) * 1.0 + scale_damp * search_factor)
    target_scale = np.maximum(0.2, np.minimum(5.0, target_scale))

    # Some book keeping
    height = self.original_target_height * target_scale
    width = self.original_target_width * target_scale
    left, up = x - get_center(bbox_width), y - get_center(bbox_height)
    self.cur_bbox = Rectangle(left, up, width, height)
    return self.cur_bbox


def run_epoch(epoch=29):

  # Checkpoint for evaluation
  CHECKPOINT_SEMANTIC = os.path.join(LOG_DIR, 'track_model_checkpoints', 'SA-Siam-Semantic-C4', 'model.ckpt-{iter_ckpt}')


  iter_ckpt = epoch * 6650 - 1
  checkpoint_semantic_path = CHECKPOINT_SEMANTIC.format(iter_ckpt=iter_ckpt)
  logging.info('Evaluating {}...'.format(checkpoint_semantic_path))

  # Read configurations from json
  model_config, _, track_config = load_cfgs(CHECKPOINT_SEMANTIC)

  track_config['log_level'] = 0  # Skip verbose logging for speed

  # Build the inference graph.
  g = tf.Graph()
  with g.as_default():
    model = inference_wrapper.InferenceWrapper()
    model.build_model(model_config, track_config)
    saver_loader_semantic   = get_saver('',  removes=[':0', '_semantic'], excepts=['appearance', 'State'])
  g.finalize()

  gpu_options = tf.GPUOptions(allow_growth=True)
  sess_config = tf.ConfigProto(gpu_options=gpu_options)

  with tf.Session(graph=g, config=sess_config) as sess:
    saver_loader_semantic.restore(sess, checkpoint_semantic_path)

    tracker = Tracker(model, model_config, track_config, sess)
    options = trax_py3.server.ServerOptions(trax_py3.region.RECTANGLE, trax_py3.image.PATH)

    # Pre init network
    tracker.init_tracker(Rectangle(0, 0, 255, 255), '/data/anfeng/tracking/code/VOT-toolkit-MORE-EPOCH/init.jpg')
    tracker.step_tracker('/data/anfeng/tracking/code/VOT-toolkit-MORE-EPOCH/init.jpg')
    with trax_py3.server.Server(options) as server:
      while True:
        tick = time.time()
        request = server.wait()
        if request.type in ["quit"]:
          break
        if request.type in ["error"]:
          break
        if request.type == "initialize":
          res = tracker.init_tracker(request.region, request.image)
          print ('Init:', request.image,request.region)
        else:
          res = tracker.step_tracker(request.image)
          print ('Track:', request.image)

        server.status(res)
        print('This loop duration is: {}s'.format(round(time.time() - tick, 2)))
if __name__ == '__main__':
  run_epoch(29)
