from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import time
import trax_py3.server
import trax_py3.region
import trax_py3.image
from trax_py3.region import Polygon
import logging
logging.getLogger().setLevel(logging.INFO)
import os
import os.path as osp
import tensorflow as tf
import numpy as np
import cv2
from cv2 import imwrite
import math

sys.path.append('/data/anfeng/tracking/code/SA-SIAM')
LOG_DIR = '/data/anfeng/tracking/code/SA-SIAM/Logs/SA-Siam'
from utils.infer_utils import convert_bbox_format
from utils.misc_utils import get_center, get, auto_select_gpu, load_cfgs
from inference import inference_wrapper

from utils.infer_utils import get_saver

class Tracker(object):
  """Tracker based on the siamese model."""

  def __init__(self, siamese_model, model_config, track_config, sess):
    self.siamese_model = siamese_model
    self.model_config = model_config
    self.track_config = track_config

    self.num_scales = track_config['num_scales']
    logging.info('track num scales -- {}'.format(self.num_scales))
    scales = np.arange(self.num_scales) - get_center(self.num_scales)
    self.search_factors = [self.track_config['scale_step'] ** x for x in scales]

    self.x_image_size = track_config['x_image_size']  # Search image size
    self.window = None  # Cosine window
    self.log_level = track_config['log_level']

    self.sess = sess

  def get_curent_region(self):
    y = self.len_edge * np.sin(self.angle_edge)
    x = self.len_edge * np.cos(self.angle_edge)
    pt0 = np.array([x[0] + x[1], y[0] + y[1]]) / 2.0
    pt1 = np.array([x[0] - x[1], y[0] - y[1]]) / 2.0
    pt2 = -pt0
    pt3 = -pt1
    logging.info('[pt0, pt1, pt2, pt3] = {}'.format([pt0, pt1, pt2, pt3]))
    return Polygon(list(map(lambda _: tuple((self.center_bbox + _).tolist()), [pt0, pt1, pt2, pt3])))

  def init_tracker(self, first_bbox, first_frame):
    # Get initial target bounding box and convert to center based
    all_pts = list(map(np.array, first_bbox.points))
    logging.info('Input Region: {}'.format(all_pts))
    assert len(all_pts) == 4
    self.center_bbox = np.mean(all_pts, axis=0)
    logging.info('self.center_bbox = {}'.format(self.center_bbox))
    center_all_edge_connect_to_pt0 = [(all_pts[0] + all_pts[i]) / 2 for i in range(1, 4)]
    logging.info('center_all_edge_connect_to_pt0 = {}'.format(center_all_edge_connect_to_pt0))
    vec_half_edge = [center_all_edge_connect_to_pt0[i] - self.center_bbox for i in range(3)]
    logging.info('vec_half_edge = {}'.format(vec_half_edge))
    len_vec = lambda a: np.sqrt(np.sum(a * a))
    is_2vec_ort = lambda a,b: np.isclose(sum((a / len_vec(a)) * (b / len_vec(b))), 0, atol=0.1)
    vec_half_edge = [_ for _ in vec_half_edge if not len_vec(_) < 0.1]
    logging.info('vec_half_edge(after no 0 vec) = {}'.format(vec_half_edge))
    assert len(vec_half_edge) == 2
    assert is_2vec_ort(vec_half_edge[0], vec_half_edge[1])
    self.len_edge   = np.array([math.sqrt(sum(_ * _)) * 2 for _ in vec_half_edge])
    self.angle_edge = np.array([math.atan2(_[1], _[0]) for _ in vec_half_edge])
    logging.info('self.len_edge = {}'.format(self.len_edge))
    logging.info('self.angle_edge = {}'.format(self.angle_edge))

    bbox_cx = self.center_bbox[0]
    bbox_cy = self.center_bbox[1]
    bbox_height = self.len_edge[0]
    bbox_width = self.len_edge[1]
    self.original_target_bbox = self.len_edge

    # Feed in the first frame image to set initial state.
    bbox_feed = [bbox_cy, bbox_cx, bbox_height, bbox_width]
    input_feed = [str(first_frame), bbox_feed]
    frame2crop_scale = self.siamese_model.initialize(self.sess, input_feed)
    cur_region = self.get_curent_region()
    logging.info('Init curent_region = {}'.format(cur_region))
    return cur_region

  def step_tracker(self, pre_frame):
    bbox_cx = self.center_bbox[0]
    bbox_cy = self.center_bbox[1]
    bbox_height = self.len_edge[0]
    bbox_width = self.len_edge[1]

    
    bbox_feed = [bbox_cy, bbox_cx, bbox_height, bbox_width]
    input_feed = [str(pre_frame), bbox_feed]

    outputs, metadata = self.siamese_model.inference_step(self.sess, input_feed)
    search_scale_list = outputs['scale_xs']
    response = outputs['response']
    response_size = response.shape[1]

    # Choose the scale whole response map has the highest peak
    if self.num_scales > 1:
      all_max_pos = []
      response_max = []
      if self.window is None:
        window = np.dot(np.expand_dims(np.hanning(response_size), 1),
                        np.expand_dims(np.hanning(response_size), 0))
        self.window = window / np.sum(window)  # normalize window
      for i_response in range(response.shape[0]):
        _response = response[i_response]

        with np.errstate(all='raise'):  # Raise error if something goes wrong
          _response = _response - np.min(_response)
          _response = _response / np.sum(_response)

        window_influence = self.track_config['window_influence']
        _response = (1 - window_influence) * _response + window_influence * self.window

        # Find maximum _response
        r_max, c_max = np.unravel_index(_response.argmax(),
                                        _response.shape)
        all_max_pos.append([r_max, c_max])
        response_max.append(response[i_response][r_max, c_max])
      response_max = np.array(response_max)

      penalties = self.track_config['scale_penalty'] * np.ones((self.num_scales))
      current_scale_idx = int(get_center(self.num_scales))
      penalties[current_scale_idx] = 1.0
      response_penalized = response_max * penalties
      best_scale = np.argmax(response_penalized)
      if np.max(response_max)<0:
        logging.warning('MAX_RESPONSE LESS THAN ZERO!')
        # best_scale = current_scale_idx
    else:
      best_scale = 0

    # Convert from crop-relative coordinates to frame coordinates
    p_coor = np.array(all_max_pos[best_scale])
    # displacement from the center in instance final representation ...
    disp_instance_final = p_coor - get_center(response_size)
    # ... in instance feature space ...
    upsample_factor = self.track_config['upsample_factor']
    disp_instance_feat = disp_instance_final / upsample_factor
    # ... Avoid empty position ...
    r_radius = int(response_size / upsample_factor / 2)
    disp_instance_feat = np.maximum(np.minimum(disp_instance_feat, r_radius), -r_radius)
    # ... in instance input ...
    disp_instance_input = disp_instance_feat * self.model_config['embed_config']['stride']
    # ... in instance original crop (in frame coordinates)
    disp_instance_frame = disp_instance_input / search_scale_list[best_scale]
    # Position within frame in frame coordinates
    y = bbox_cy
    x = bbox_cx
    y += disp_instance_frame[0]
    x += disp_instance_frame[1]

    self.center_bbox[0] = x
    self.center_bbox[1] = y

    # Target scale damping and saturation
    target_scale = self.len_edge[0] / self.original_target_bbox[0]
    search_factor = self.search_factors[best_scale]
    scale_damp = self.track_config['scale_damp']  # damping factor for scale update
    target_scale *= ((1 - scale_damp) * 1.0 + scale_damp * search_factor)
    target_scale = np.maximum(0.2, np.minimum(5.0, target_scale))

    self.len_edge = self.original_target_bbox * target_scale
    logging.info('self.center_bbox = {}'.format(self.center_bbox))
    logging.info('self.len_edge = {}'.format(self.len_edge))
    # Some book keeping
    cur_region = self.get_curent_region()
    logging.info('Step curent_region = {}'.format(cur_region))
    return cur_region


def run_epoch(epoch=29, scale_penalty=None, windows_inf=None):

  # Checkpoint for evaluation
  CHECKPOINT_APPEARANCE = os.path.join(LOG_DIR, 'track_model_checkpoints', 'SA-Siam-Appearance', 'model.ckpt-{iter_ckpt}')
  CHECKPOINT_SEMANTIC = os.path.join(LOG_DIR, 'track_model_checkpoints', 'SA-Siam-Semantic', 'model.ckpt-{iter_ckpt}')
  CHECKPOINT_SA_SIAM = os.path.join(LOG_DIR, 'track_model_checkpoints', 'SA-Siam')


  iter_ckpt = epoch * 6650 - 1
  checkpoint_appearance_path = CHECKPOINT_APPEARANCE.format(iter_ckpt=iter_ckpt)
  logging.info('Evaluating {}...'.format(checkpoint_appearance_path))
  checkpoint_semantic_path = CHECKPOINT_SEMANTIC.format(iter_ckpt=iter_ckpt)
  logging.info('Evaluating {}...'.format(checkpoint_semantic_path))

  # Read configurations from json
  model_config, _, track_config = load_cfgs(CHECKPOINT_SA_SIAM)
  if scale_penalty is not None:
    track_config['scale_penalty'] = scale_penalty
  if windows_inf is not None:
    track_config['window_influence'] = windows_inf
  track_config['log_level'] = 0  # Skip verbose logging for speed
  model_config['sa_siam_config']['en_semantic'] = False

  # Build the inference graph.
  g = tf.Graph()
  with g.as_default():
    model = inference_wrapper.InferenceWrapper()
    model.build_model(model_config, track_config)
    saver_loader_semantic   = get_saver('',  removes=[':0', '_semantic'], excepts=['appearance', 'State'])
    saver_loader_appearance = get_saver('',  removes=[':0', '_appearance'], excepts=['semantic',  'State'])
  g.finalize()

  gpu_options = tf.GPUOptions(allow_growth=True)
  sess_config = tf.ConfigProto(gpu_options=gpu_options)

  with tf.Session(graph=g, config=sess_config) as sess:
    saver_loader_semantic.restore(sess, checkpoint_semantic_path)
    saver_loader_appearance.restore(sess, checkpoint_appearance_path)

    tracker = Tracker(model, model_config, track_config, sess)
    options = trax_py3.server.ServerOptions(trax_py3.region.POLYGON, trax_py3.image.PATH)

    # Pre init network
    tracker.init_tracker(Polygon([(0, 0), (0, 255), (255, 255), (255,0)]), '/data/anfeng/tracking/code/VOT-toolkit-MORE-EPOCH/init.jpg')
    tracker.step_tracker('/data/anfeng/tracking/code/VOT-toolkit-MORE-EPOCH/init.jpg')
    with trax_py3.server.Server(options) as server:
      while True:
        tick = time.time()
        request = server.wait()
        if request.type in ["quit"]:
          break
        if request.type in ["error"]:
          break
        if request.type == "initialize":
          res = tracker.init_tracker(request.region, request.image)
          print ('Init:', request.image,request.region)
        else:
          res = tracker.step_tracker(request.image)
          print ('Track:', request.image)

        server.status(res)
        print('This loop duration is: {}s'.format(round(time.time() - tick, 2)))
if __name__ == '__main__':
  run_epoch(29)
